TerraLifePlusDevTools = {}
local TerraLifePlusDevTools_mt = Class(TerraLifePlusDevTools)

function TerraLifePlusDevTools.new()
	local self = setmetatable({}, TerraLifePlusDevTools_mt)
	addConsoleCommand("tlpMowField", "Mows given field. You need to sit in a Big M to execute this.", "debugMowField", self)
	addConsoleCommand("tlpSpawnAllGrowthStates", "Spawns given fruit type index in up to 32 fields, where each field represents one growth state", "debugSpawnAllGrowthStates", self)
	addConsoleCommand("tlpMulchField", "Mulches given field.", "debugMulchField", self)
	addConsoleCommand("tlpTestGrassYields", "Tests the yields of all grass types", "debugTestGrassYields", self)
	return self
end

function TerraLifePlusDevTools:deleteMap()
	removeConsoleCommand("tlpMowField")
	removeConsoleCommand("tlpSpawnAllGrowthStates")
end

function TerraLifePlusDevTools:getCoordsFromFieldDimensions(fieldDimensions, index)
	local dimWidth = getChildAt(fieldDimensions, index)
	local dimStart = getChildAt(dimWidth, 0)
	local dimHeight = getChildAt(dimWidth, 1)
	local startWorldX, _, startWorldZ = getWorldTranslation(dimStart)
	local widthWorldX, _, widthWorldZ = getWorldTranslation(dimWidth)
	local heightWorldX, _, heightWorldZ = getWorldTranslation(dimHeight)

	return startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ, dimStart, dimWidth, dimHeight
end

function TerraLifePlusDevTools:debugMowField(fieldIndexString)
	if tonumber(fieldIndexString) == nil then
		Logging.error("Please enter a valid fruit type index")
		return
	end
	local fieldIndex = tonumber(fieldIndexString)
	local currentVehicle = g_currentMission.controlledVehicle
	if currentVehicle == nil or currentVehicle.spec_mower == nil then
		Logging.error("You need to sit in a mowing vehicle in order to use this command")
		return
	end
	local field = g_fieldManager:getFieldByIndex(tonumber(fieldIndex))
	if field ~= nil and field.fieldDimensions ~= nil then
		local dimensionCount = getNumOfChildren(field.fieldDimensions)
		for inputFruitType, converterData in pairs(currentVehicle.spec_mower.fruitTypeConverters) do
			local totalLiters = 0
			local totalSqm = 0
			for i = 1, dimensionCount do
				local startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ, start, width, height
					= self:getCoordsFromFieldDimensions(field.fieldDimensions, i - 1)
				local workAreaForPf = {
					start = start,
					width = width,
					height = height
				}
				local harvestExtension = FS22_precisionFarming.g_precisionFarming.harvestExtension
				harvestExtension:preProcessMowerArea(currentVehicle, workAreaForPf, 0)
				local changedArea, totalArea, sprayFactor, plowFactor, limeFactor, weedFactor, stubbleFactor, rollerFactor, _, growthState, _ 
					= FSDensityMapUtil.updateMowerArea(inputFruitType, startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ, false)

				if changedArea > 0 then
					-- Factors for the harvest scale multiplier are outdated if we don't call this here for some reason
					harvestExtension:preProcessMowerArea(currentVehicle, workAreaForPf, 0)
					local multiplier = g_currentMission:getHarvestScaleMultiplier(inputFruitType, sprayFactor, plowFactor, limeFactor, weedFactor, stubbleFactor, rollerFactor)
					changedArea = changedArea * multiplier
					local pixelToSqm = g_currentMission:getFruitPixelsToSqm()
					local sqm = changedArea * pixelToSqm
					local litersToDrop = sqm * TLMowerExtension.getFillTypeLiterPerSqm(inputFruitType, converterData.fillTypeIndex)
					print(("multiplier=%s, changedArea=%s, pixelToSqm=%s, sqm=%s, litersToDrop=%s"):format(multiplier, changedArea/multiplier, pixelToSqm, sqm, litersToDrop))
					local xDiff = (widthWorldX - startWorldX) / 2
					local zDiff = (widthWorldZ - startWorldZ) / 2
					local sx = startWorldX + xDiff
					local sz = startWorldZ + zDiff
					local ex = heightWorldX + xDiff
					local ez = heightWorldZ + zDiff
					local sy = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, sx, 0, sz)
					local ey = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, ex, 0, ez)
					DensityMapHeightUtil.tipToGroundAroundLine(self, litersToDrop, converterData.fillTypeIndex, sx, sy, sz, ex, ey, ez, 0, nil, 0, false, nil, false)
					totalLiters = totalLiters + litersToDrop
					totalSqm = totalSqm + (sqm / multiplier)
				end
			end
			if totalLiters > 0 then
				local ha = totalSqm / 10000
				local litersPerHa = totalLiters / ha
				print(("[TLP DEV] Dropped %d liters of %s on %.3f ha, resulting in %d l / ha"):format(totalLiters, g_fillTypeManager:getFillTypeNameByIndex(converterData.fillTypeIndex), ha, litersPerHa))
			end
		end
	else
		Logging.error("Invalid field, or no field dimensions found")
	end
end

function TerraLifePlusDevTools:debugMulchField(fieldIndexString)
	if tonumber(fieldIndexString) == nil then
		Logging.error("Please enter a valid fruit type index")
		return
	end
	local fieldIndex = tonumber(fieldIndexString)
	local field = g_fieldManager:getFieldByIndex(tonumber(fieldIndex))
	if field ~= nil and field.fieldDimensions ~= nil then
		local dimensionCount = getNumOfChildren(field.fieldDimensions)
		for i = 1, dimensionCount do
			local startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ, start, width, height
				= self:getCoordsFromFieldDimensions(field.fieldDimensions, i - 1)
			FSDensityMapUtil.updateMulcherArea(startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ)
		end
	else
		Logging.error("Invalid field, or no field dimensions found")
	end
end

function TerraLifePlusDevTools:debugSpawnAllGrowthStates(fruitTypeIndexString)
	if tonumber(fruitTypeIndexString) == nil then
		Logging.error("Please enter a valid fruit type index")

		local sortedFruitTypes = {table.unpack(g_fruitTypeManager:getFruitTypes())}
		table.sort(sortedFruitTypes, function(a, b) return a.name < b.name end)
		for _, fruitType in ipairs(sortedFruitTypes) do
			print(string.format("%s - %s", fruitType.name, fruitType.index))
		end

		return
	end
	local fruitTypeIndex = tonumber(fruitTypeIndexString)
	if fruitTypeIndex < 0 or fruitTypeIndex > #g_fruitTypeManager:getFruitTypes() then
		Logging.error("Please enter a valid fruit type index")
		return
	end
	local easyDev = g_globalMods.easyDevControls
	if not easyDev then
		Logging.error("Easy Dev Controls not detected")
		return
	end
	if not easyDev.setFieldFruit then
		Logging.error("Easy Dev Controls setFieldFruit function not detected")
		return
	end
	local fruitDesc = g_fruitTypeManager:getFruitTypeByIndex(fruitTypeIndex)

	-- Remember which growth states are supported by the fruit (some of them may have gaps)
	local growthStates = {}
	for i = 1, fruitDesc.numGrowthStates do
		growthStates[i] = true
	end
	if fruitDesc.witheredState ~= nil then
		growthStates[fruitDesc.witheredState] = true
		print(("Field %d will have a withered state"):format(fruitDesc.witheredState))
	end
	if fruitDesc.destruction.state ~= nil then
		growthStates[fruitDesc.cutState] = true
		print(("Field %d will have a cut state"):format(fruitDesc.cutState))
	end
	if fruitDesc.mulcher.state ~= nil then
		growthStates[fruitDesc.mulcher.state] = true
		print(("Field %d will have a mulched state"):format(fruitDesc.mulcher.state))
	end

	-- Spawn each growth state on the field, reset the field if there is no fruit
	for growthState = 1, 31 do
		local fieldFruitState = 0
		if growthStates[growthState] then
			fieldFruitState = growthState
		end

		local mulchField = false
		if fruitDesc.mulcher.state ~= nil and growthState == fruitDesc.mulcher.state then
			-- Additionally mulch the field
			mulchField = true
			-- Also use the cut state, if available, so we simulate mulching a harvested field which is what most players would do
			fieldFruitState = fruitDesc.cutState or fieldFruitState
		end

		-- Spawn fruit in this growth state
		local fieldIndex = growthState
		local stoneMinValue, _ = g_currentMission.stoneSystem:getMinMaxValues()
		easyDev:setFieldFruit(
			fieldIndex, fruitTypeIndex, fieldFruitState,
			1, -- ground layer
			0, -- not fertilized (unused with PF)
			1, -- plowed
			1, -- limed
			1, -- mulching bonus
			0, -- no weeds
			0, -- no herbicide sprayed 
			1, -- rolling bonus
			stoneMinValue,
			true, -- buy farmland
			g_currentMission:getFarmId()
		)
		if mulchField then
			self:debugMulchField(fieldIndex)
		end
	end
end

function TerraLifePlusDevTools:debugTestGrassYields()
	local GRASS_TYPES = {
		"CLOVERGRASS",
		"GRASS",
		"MEADOW",
		"HORSEGRASS",
		"FIELDGRASS",
		"HERBGRASS",
		"PASTUREGRASS",
		"LANDSCAPEGRASS",
		"ROADSIDEGRASS",
		"ORCHARDGRASS",
		"LAWNGRASS",
		"ALFALFA",
		"CLOVER"
	}

	local yields = {}
	local maxPfModifier = 2.5 -- + 100% for perfectly prepared field / * 125% for best soil type
	local oneHectar = 10000 -- m²
	for _, grassTypeName in ipairs(GRASS_TYPES) do
		local fruitDesc = g_fruitTypeManager:getFruitTypeByName(grassTypeName)
		if fruitDesc == nil then
			Logging.error(("Could not find grass type %s"):format(grassTypeName))
			return
		end
		local currentYields = {}
		currentYields.name = grassTypeName
		-- Calculate yields for mowers without conditioners + forage wagons:
		local converterData = g_fruitTypeManager:getConverterDataByName("MOWER")
		local converter = converterData[fruitDesc.index]
		if converter == nil then
			Logging.error("No MOWER converter found for %s", grassTypeName)
			return
		end
		local targetFillTypeIndex = converter.fillTypeIndex
		local unknownFactor = 1.05 -- In tests, yields were always about 5% higher than calculated, not sure why
		currentYields.mower = oneHectar * maxPfModifier * unknownFactor * TLMowerExtension.getFillTypeLiterPerSqm(fruitDesc.index, targetFillTypeIndex)

		-- Calculate yields for mowers with conditioners + forage wagons:
		converterData = g_fruitTypeManager:getConverterDataByName("MOWERCONDITIONER")
		local converter = converterData[fruitDesc.index]
		if converter == nil then
			Logging.error("No MOWERCONDITIONER converter found for %s", grassTypeName)
			return
		end
		targetFillTypeIndex = converter.fillTypeIndex
		currentYields.mowerConditioner = oneHectar * maxPfModifier * unknownFactor * TLMowerExtension.getFillTypeLiterPerSqm(fruitDesc.index, targetFillTypeIndex)

		-- Calculate yields for mowers with conditioners + forage harvesters with pickup:
		-- Two step process: Mow as conditioned first, then pick up
		-- There is however a bug in the Cutter:processPickupCutterArea implementation where the conversion factor of the first fruit type with this output fill type is being used
		converterData = g_fruitTypeManager:getConverterDataByName("FORAGEHARVESTER")
		local conversionFactor = 0
		for _, tempFruitDesc in ipairs(g_fruitTypeManager:getFruitTypes()) do
			if converterData[tempFruitDesc.index] ~= nil then
				-- Find the conversion factor exactly like the base game does:
				if converterData[tempFruitDesc.index].fillTypeIndex == targetFillTypeIndex then
					conversionFactor = converterData[tempFruitDesc.index].conversionFactor
					print(("Using conversion factor %.3f of Fruit Type %s and Fill Type Index % for Grass Type %s"):format(conversionFactor, tempFruitDesc.name, targetFillTypeIndex, grassTypeName))
					break
				end
			end
		end
		if conversionFactor == 0 then
			local fillTypeName = g_fillTypeManager:getFillTypeNameByIndex(targetFillTypeIndex)
			Logging.warning("No FORAGEHARVESTER converter found for fruit type %s and fill type %s", grassTypeName, fillTypeName)
			conversionFactor = 1
		end
		unknownFactor = 0.95 -- In tests, yield was always roughly 5% below the calculated values
		currentYields.forageHarvesterPickup = currentYields.mowerConditioner * conversionFactor * unknownFactor

		-- Calculate yields for forage harvesters with mower:
		-- Note. In this case, the proper conversion factor is being used, but also the fill type's literPerSqm rather than the windrow's litersPerSqm
		conversionFactor = 1
		if converterData[fruitDesc.index] ~= nil then
			conversionFactor = converterData[fruitDesc.index].conversionFactor
		else
			Logging.warning("No FORAGEHARVESTER converter found at all for fruit type %s", grassTypeName)
		end
		unknownFactor = 1.15 -- For Forage Harvesters with cutters, the test results were always roughly 15% above the calculated values
		currentYields.forageHarvesterCutter = oneHectar * maxPfModifier * conversionFactor * fruitDesc.literPerSqm * unknownFactor

		table.insert(yields, currentYields)
	end

	print("----------------------------------------------")
	print("Current grass Yields")
	print("Fruit Type, Mower Yield, MowerConditioner Yield, ForagePickup Yield, ForageCutter Yield")
	for _, data in ipairs(yields) do
		print(("%s, %d, %d, %d, %d"):format(data.name, data.mower, data.mowerConditioner, data.forageHarvesterPickup, data.forageHarvesterCutter or 0))
	end
	print("----------------------------------------------")

end


g_terraLifePlusDevTools = TerraLifePlusDevTools.new()
addModEventListener(g_terraLifePlusDevTools)