TLPlayerHUDInfoBoxExtension = {}

local function getTargetWeedStateInfo(weedStateInfo, isColorBlindMode)
	if weedStateInfo == nil then
		-- Desired tool is ineffective
		return { name = g_i18n:getText("TL_weed_ineffective"), color = isColorBlindMode and TerraLiveExtension.BLINDCOLORS.BAD or TerraLiveExtension.COLORS.BAD }
	elseif not weedStateInfo.hasGrowth and not weedStateInfo.canBeWeeded and not weedStateInfo.canBeHoed and not weedStateInfo.canBeSprayed then
		-- dead states
		return { name = g_i18n:getText("TL_weed_effective"), color = isColorBlindMode and TerraLiveExtension.BLINDCOLORS.GOOD or TerraLiveExtension.COLORS.GOOD }
	else
		-- Something can be done about this state -> not fully gone
		return { name = g_i18n:getText("TL_weed_partially_effective"), color = isColorBlindMode and TerraLiveExtension.BLINDCOLORS.NEUTRAL or TerraLiveExtension.COLORS.NEUTRAL }
	end
end

local function getInfoAboutTargetWeedState(index, weedStateInfos)
	if index ~= nil then
		return getTargetWeedStateInfo(weedStateInfos[index])
	else
		return getTargetWeedStateInfo(nil)
	end
end

local function addToolEffectivenessLine(box, toolName, targetInfo, isEffective, ineffectiveInfo)
	if isEffective then
		box:addLine(g_i18n:getText(toolName), targetInfo.name, true, targetInfo.color)
	else
		box:addLine(g_i18n:getText(toolName), ineffectiveInfo.name, true, ineffectiveInfo.color)
	end
end

-- Build lookup tables for the kind of effects each tool would have on weeds
function TLPlayerHUDInfoBoxExtension.loadWeed(weedSystem, ...)
	local maxWeedState = 63

	-- Build a lookup table for fast access later
	local weedStateInfos = {}
	for i = 0, maxWeedState do
		local data = {}

		data.canBeMulched = weedSystem.mulcherReplacements.weed.replacements[i] ~= nil
		data.canBeWeeded = weedSystem.weederReplacements.weed.replacements[i] ~= nil
		data.canBeHoed = weedSystem.weederHoeReplacements.weed.replacements[i] ~= nil
		data.canBeSprayed = weedSystem.herbicideReplacements.weed.replacements[i] ~= nil
		data.hasGrowth = weedSystem.growthMapping[i] ~= nil
		data.weedFactor = weedSystem.factors[i] or 0

		weedStateInfos[i] = data
	end

	-- Remember the expected effectiveness for each tool when applied to each weed state
	-- (allows much faster lookup when drawing the HUD)
	local isColorBlindMode = g_gameSettings:getValue(GameSettings.SETTING.USE_COLORBLIND_MODE) or false
	weedSystem.tlpWeedStateInfo = {}
	for i = 1, maxWeedState do
		weedSystem.tlpWeedStateInfo[i] = {
			ownName = weedSystem.fieldInfoStates[i],
			hasGrowth = weedStateInfos[i].hasGrowth,
			weedFactor = weedStateInfos[i].weedFactor,
			weederTargetInfo = getInfoAboutTargetWeedState(weedSystem.weederReplacements.weed.replacements[i], weedStateInfos, isColorBlindMode),
			hoeTargetInfo = getInfoAboutTargetWeedState(weedSystem.weederHoeReplacements.weed.replacements[i], weedStateInfos, isColorBlindMode),
			herbicideTargetInfo = getInfoAboutTargetWeedState(weedSystem.herbicideReplacements.weed.replacements[i], weedStateInfos, isColorBlindMode)
		}
	end
end

-- Replace the base game information about weeds in order to handle our extended weed
function TLPlayerHUDInfoBoxExtension.fieldAddWeed(playerHudUpdater, superFunc, data, box)
	if g_currentMission.missionInfo.weedsEnabled then
		local weedSystem = g_currentMission.weedSystem
		local fieldInfoStates = weedSystem:getFieldInfoStates()
		local maxState = nil
		local maxPixels = 0

		for state, pixels in pairs(data.weed) do
			if maxPixels < pixels then
				maxState = state
				maxPixels = pixels
			end
		end

		if maxState == nil then
			return
		end

		local fruitTypeIndex = data.fruitTypeMax
		local fruitGrowthState = data.fruitStateMax
		local fruitTypeDesc = nil

		if fruitTypeIndex ~= nil then
			fruitTypeDesc = g_fruitTypeManager:getFruitTypeByIndex(fruitTypeIndex)
		end

		local weedStateInfo = weedSystem.tlpWeedStateInfo[maxState]
		if not weedStateInfo then
			Logging.warning("Weed Treatment targets are empty for state %d", maxState)
			return
		end

		if fruitTypeDesc == nil then
			return -- happens with bushes etc. No point in displaying any infos there
		end

		if g_showDevelopmentWarnings then
			box:addLine("DEBUG: Weed State", tostring(maxState))
			box:addLine("DEBUG: Too Late?", (fruitGrowthState > fruitTypeDesc.maxWeederState) and "yes" or "no")
			box:addLine("DEBUG: Has growth?", weedStateInfo.hasGrowth and "yes" or "no")
			box:addLine("DEBUG: Weed Factor", tostring(weedStateInfo.weedFactor))
		end

		if weedStateInfo.ownName then
			box:addLine(g_i18n:getText("TL_weed_state"), weedStateInfo.ownName, true)

			local ineffectiveInfo = getTargetWeedStateInfo(nil)

			local weederIsEffective = fruitTypeIndex == nil or fruitGrowthState < fruitTypeDesc.minHarvestingGrowthState and fruitGrowthState <= fruitTypeDesc.maxWeederState
			addToolEffectivenessLine(box, "weed_destruction_weeder", weedStateInfo.weederTargetInfo, weederIsEffective, ineffectiveInfo)

			local hoeIsEffective = fruitTypeIndex == nil or fruitGrowthState < fruitTypeDesc.minHarvestingGrowthState and fruitGrowthState <= fruitTypeDesc.maxWeederHoeState
			addToolEffectivenessLine(box, "weed_destruction_hoe", weedStateInfo.hoeTargetInfo, hoeIsEffective, ineffectiveInfo)

			local herbicideIsEffective = fruitTypeIndex == nil or fruitGrowthState < fruitTypeDesc.minHarvestingGrowthState
			addToolEffectivenessLine(box, "weed_destruction_herbicide", weedStateInfo.herbicideTargetInfo, herbicideIsEffective, ineffectiveInfo)
		-- else: dead weeds. Shouldn't be called in this case anyway
		end

	end
end

WeedSystem.loadWeed = Utils.appendedFunction(WeedSystem.loadWeed, TLPlayerHUDInfoBoxExtension.loadWeed)
PlayerHUDUpdater.fieldAddWeed = Utils.overwrittenFunction(PlayerHUDUpdater.fieldAddWeed, TLPlayerHUDInfoBoxExtension.fieldAddWeed)