-- Farming Agency Multi Item Purchase
-- Allows purchasing of different Pallet-Types, Configurations and FillLevels on one Shop-Screen
-- by [F/A] Braeven
-- Version 1.0 - 10.09.2023
-- Version 1.1 - 25.09.2023 Made more flexible
-- Version 1.2 - 10.01.2024 Added Bales

-- Only for use with TerraLife, it is not allowed to use this script without permission from Braeven / Farming Agency


FAMultiItemPurchase = {
	initSpecialization = function ()
		g_configurationManager:addConfigurationType("fAPalletFillTypes", g_i18n:getText("TL_FAMultiItemPurchase_FillType"), nil, nil, nil, nil, ConfigurationUtil.SELECTOR_MULTIOPTION)
		g_configurationManager:addConfigurationType("fAPalletFillLevels", g_i18n:getText("TL_FAMultiItemPurchase_Packaging"), nil, nil, nil, nil, ConfigurationUtil.SELECTOR_MULTIOPTION)
		g_configurationManager:addConfigurationType("fAPalletAmount", g_i18n:getText("TL_FAMultiItemPurchase_Amount"), nil, nil, nil, nil, ConfigurationUtil.SELECTOR_MULTIOPTION)

		local schema = Vehicle.xmlSchema

		schema:setXMLSpecializationType("FAMultiItemPurchase")
		schema:register(XMLValueType.STRING, "vehicle.fAPalletFillTypesConfigurations.fAPalletFillTypesConfiguration(?).option(?)#filename", "pallet filename")
		schema:register(XMLValueType.STRING, "vehicle.fAPalletFillTypesConfigurations.fAPalletFillTypesConfiguration(?).option(?)#fillUnit", "FillUnit")
		schema:register(XMLValueType.STRING, "vehicle.fAPalletFillTypesConfigurations.fAPalletFillTypesConfiguration(?).option(?)#design", "design ID")
		schema:register(XMLValueType.BOOL, "vehicle.fAPalletFillTypesConfigurations.fAPalletFillTypesConfiguration(?).option(?)#isBale", "Is object a bale")
		schema:register(XMLValueType.BOOL, "vehicle.fAPalletFillTypesConfigurations.fAPalletFillTypesConfiguration(?).option(?)#isWrapped", "Is bale wrapped")

		schema:register(XMLValueType.VECTOR_TRANS, "vehicle.fAPalletAmount.itemPositions(?).itemPosition(?)#position", "pallet/bale position")
		schema:register(XMLValueType.VECTOR_ROT, "vehicle.fAPalletAmount.itemPositions(?).itemPosition(?)#rotation", "pallet/bale rotation")
		
		schema:register(XMLValueType.STRING, "vehicle.fAPalletAmountConfigurations.fAPalletAmountConfiguration(?)#discount", "Discount on buying more than one pallet")
		schema:register(XMLValueType.STRING, "vehicle.fAPalletFillLevelsConfigurations.fAPalletFillLevelsConfiguration(?)#surcharge", "Discount on buying not full")
		schema:register(XMLValueType.STRING, "vehicle.fAPalletFillTypesConfigurations.fAPalletFillTypesConfiguration(?)#fillType", "FillType-Name")

		schema:setXMLSpecializationType(XMLManager.XML_SPECIALIZATION_NONE)
	end,
	prerequisitesPresent = function (specializations)
		return SpecializationUtil.hasSpecialization(FillUnit, specializations)
	end
}

function FAMultiItemPurchase.registerFunctions(vehicleType)
	SpecializationUtil.registerFunction(vehicleType, "loadItemAtPosition", FAMultiItemPurchase.loadItemAtPosition)
	SpecializationUtil.registerFunction(vehicleType, "onFinishLoadingVehicle", FAMultiItemPurchase.onFinishLoadingVehicle)
end

function FAMultiItemPurchase.registerOverwrittenFunctions(vehicleType)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "getTotalMass", FAMultiItemPurchase.getTotalMass)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "getFillUnitCapacity", FAMultiItemPurchase.getFillUnitCapacity)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "setVisibility", FAMultiItemPurchase.setVisibility)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "addToPhysics", FAMultiItemPurchase.addToPhysics)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "removeFromPhysics", FAMultiItemPurchase.removeFromPhysics)
end

function FAMultiItemPurchase.registerEventListeners(vehicleType)
	SpecializationUtil.registerEventListener(vehicleType, "onLoad", FAMultiItemPurchase)
	SpecializationUtil.registerEventListener(vehicleType, "onPreLoadFinished", FAMultiItemPurchase)
	SpecializationUtil.registerEventListener(vehicleType, "onDelete", FAMultiItemPurchase)
	SpecializationUtil.registerEventListener(vehicleType, "onUpdate", FAMultiItemPurchase)
end

function FAMultiItemPurchase:onLoad(savegame)
	if self.spec_fAMultiItemPurchase == nil then
		self.spec_fAMultiItemPurchase = {}
	end
	local spec = self.spec_fAMultiItemPurchase
	spec.loadedVehicles = {}
	spec.loadedBales = {}
	spec.filenames = {}
	spec.fillUnits = {}
	spec.fillTypes = {}
	spec.designs = {}
	spec.isBaleList = {}
	spec.isWrappedList = {}

	self.xmlFile:iterate("vehicle.fAPalletFillTypesConfigurations.fAPalletFillTypesConfiguration", function(i, baseKey)
		local filenames = {}
		local fillUnits = {}
		local designs = {}
		local isBale = {}
		local isWrapped = {}

		self.xmlFile:iterate(baseKey..".option", function(x, optionKey)
			local filename = Utils.getFilename(self.xmlFile:getValue(optionKey .. "#filename", nil, true), self.baseDirectory)
			filenames[x] = filename
			local fillUnit = tonumber(self.xmlFile:getValue(optionKey .. "#fillUnit", nil, true))
			fillUnits[x] = fillUnit
			local design = tonumber(self.xmlFile:getValue(optionKey .. "#design", nil, true))
			designs[x] = design
			isBale[x] = self.xmlFile:getValue(optionKey .. "#isBale", false)
			isWrapped[x] = self.xmlFile:getValue(optionKey .. "#isWrapped", false)
		end)

		spec.fillTypes[i] = g_fillTypeManager:getFillTypeIndexByName(self.xmlFile:getValue(baseKey .. "#fillType", nil, true))
		spec.filenames[i] = filenames
		spec.fillUnits[i] = fillUnits
		spec.designs[i] = designs
		spec.isBaleList[i] = isBale
		spec.isWrappedList[i] = isWrapped
	end)

	spec.positions = {}
	spec.rotations = {}

	self.xmlFile:iterate("vehicle.fAPalletAmount.itemPositions", function(i, baseKey)
		local positions = {}
		local rotations = {}

		self.xmlFile:iterate(baseKey..".itemPosition", function(x, itemKey)
			local position = self.xmlFile:getValue(itemKey .. "#position", nil, true)
			local rotation = self.xmlFile:getValue(itemKey .. "#rotation", nil, true)
			positions[x] = position
			rotations[x] = rotation
		end)

		spec.positions[i] = positions
		spec.rotations[i] = rotations
	end)
	
	if not self.isServer then
		SpecializationUtil.removeEventListener(self, "onUpdate", FAMultiItemPurchase)
	end
end

function FAMultiItemPurchase:onPreLoadFinished(savegame)
	local spec = self.spec_fAMultiItemPurchase
	spec.fillUnit = spec.fillUnits[self.configurations.fAPalletFillTypes]
	spec.design = spec.designs[self.configurations.fAPalletFillTypes]
	spec.filename = spec.filenames[self.configurations.fAPalletFillTypes]
	spec.fillType = spec.fillTypes[self.configurations.fAPalletFillTypes]
	spec.isBale = spec.isBaleList[self.configurations.fAPalletFillTypes]
	spec.isWrapped = spec.isWrappedList[self.configurations.fAPalletFillTypes]
	local rotations = spec.rotations[self.configurations.fAPalletFillLevels]
	local positions = spec.positions[self.configurations.fAPalletFillLevels]
	
	for x=1, self.configurations.fAPalletAmount do
		local position = {}
		position.rotation = rotations[x]
		position.position = positions[x]

		self:loadItemAtPosition(position)
	end
end

function FAMultiItemPurchase:loadItemAtPosition(position)
	local spec = self.spec_fAMultiItemPurchase

	if self.isServer or self.propertyState == Vehicle.PROPERTY_STATE_SHOP_CONFIG then
		local x, y, z = localToWorld(self.components[1].node, unpack(position.position))
		local rx, ry, rz = localRotationToWorld(self.components[1].node, unpack(position.rotation))

		if spec.isBale[self.configurations.fAPalletFillLevels] == true then
			local baleObject = Bale.new(self.isServer, self.isClient)
			if baleObject:loadFromConfigXML(spec.filename[self.configurations.fAPalletFillLevels], x, y, z, rx, ry, rz) then
				baleObject:setFillType(spec.fillType, true)
				baleObject:setOwnerFarmId(self:getActiveFarm(), true)

				if self.propertyState ~= Vehicle.PROPERTY_STATE_SHOP_CONFIG then
					baleObject:register()
				end

				if spec.isWrapped[self.configurations.fAPalletFillLevels] == true then
					baleObject:setWrappingState(1)

					if self.configurations.baseColor ~= nil then
						local color = ConfigurationUtil.getColorByConfigId(self, "baseColor", self.configurations.baseColor)

						baleObject:setColor(unpack(color))
					end
				end

				setPairCollision(self.components[1].node, baleObject.nodeId, false)
				table.insert(spec.loadedBales, baleObject)
			end
		else
			local location = {
				x = x,
				y = y,
				z = z,
				xRot = rx,
				yRot = ry,
				zRot = rz
			}
			self.subLoadingTasksFinished = false
			self.numPendingSubLoadingTasks = self.numPendingSubLoadingTasks + 1
			local registerVehicle = self.propertyState ~= Vehicle.PROPERTY_STATE_SHOP_CONFIG
			local forceServer = self.propertyState == Vehicle.PROPERTY_STATE_SHOP_CONFIG
			local file = spec.filename[self.configurations.fAPalletFillLevels]
			local config = {}
			if spec.design[self.configurations.fAPalletFillLevels] ~= 0 then
				config["design"] = spec.design[self.configurations.fAPalletFillLevels]
			end
			if spec.fillUnit[self.configurations.fAPalletFillLevels] ~= 0 then
				config["fillUnit"] = spec.fillUnit[self.configurations.fAPalletFillLevels]
			end

			VehicleLoadingUtil.loadVehicle(file, location, true, 0, self.propertyState, self:getActiveFarm(), config, nil, self.onFinishLoadingVehicle, self, nil, registerVehicle, forceServer)
		end
	end
end

function FAMultiItemPurchase:onFinishLoadingVehicle(vehicle, vehicleLoadState)
	local spec = self.spec_fAMultiItemPurchase

	if vehicleLoadState == VehicleLoadingUtil.VEHICLE_LOAD_OK then
		table.insert(spec.loadedVehicles, vehicle)
	end

	self.numPendingSubLoadingTasks = self.numPendingSubLoadingTasks - 1
	self.subLoadingTasksFinished = self.numPendingSubLoadingTasks == 0

	if not self.isDeleted and not self.isDeleting and self.syncVehicleLoadingFinished then
		self:tryFinishLoading()
	end
end

function FAMultiItemPurchase:onDelete()
	if self.propertyState == Vehicle.PROPERTY_STATE_SHOP_CONFIG or g_iconGenerator ~= nil or g_currentMission.debugVehiclesToBeLoaded ~= nil then
		local spec = self.spec_fAMultiItemPurchase

		if spec.loadedVehicles ~= nil then
			for _, vehicle in ipairs(spec.loadedVehicles) do
				vehicle:delete()
			end
		end
		if spec.loadedBales ~= nil then
			for _, bale in ipairs(spec.loadedBales) do
				bale:delete()
			end
		end
	end
end

function FAMultiItemPurchase:onUpdate(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
	if self.propertyState ~= Vehicle.PROPERTY_STATE_SHOP_CONFIG then
		g_currentMission:removeVehicle(self)
	end
end

function FAMultiItemPurchase:getTotalMass(superFunc, onlyGivenVehicle)
	if self.propertyState == Vehicle.PROPERTY_STATE_SHOP_CONFIG then
		local mass = 0
		local spec = self.spec_fAMultiItemPurchase

		for _, vehicle in ipairs(spec.loadedVehicles) do
			mass = mass + vehicle:getTotalMass()
		end

		for _, bale in ipairs(spec.loadedBales) do
			mass = mass + bale:getMass()
		end

		return mass
	end

	return 0
end

function FAMultiItemPurchase:getFillUnitCapacity(superFunc, fillUnitIndex)
	if self.propertyState == Vehicle.PROPERTY_STATE_SHOP_CONFIG then
		local fillLevel = 0
		local spec = self.spec_fAMultiItemPurchase

		for _, bale in ipairs(spec.loadedBales) do
			fillLevel = fillLevel + bale:getFillLevel()
		end

		for _, vehicle in ipairs(spec.loadedVehicles) do
			if vehicle.getFillUnitCapacity ~= nil then
				fillLevel = fillLevel + vehicle:getFillUnitCapacity(1)
			end
		end

		return fillLevel
	end

	return 0
end

function FAMultiItemPurchase:setVisibility(superFunc, state)
	local spec = self.spec_fAMultiItemPurchase

	for _, vehicle in ipairs(spec.loadedVehicles) do
		vehicle:setVisibility(state)
	end

	superFunc(self, state)
end

function FAMultiItemPurchase:addToPhysics(superFunc)
end

function FAMultiItemPurchase:removeFromPhysics(superFunc)
	local spec = self.spec_fAMultiItemPurchase

	for _, bale in ipairs(spec.loadedBales) do
		bale:removeFromPhysics()
	end

	for _, vehicle in ipairs(spec.loadedVehicles) do
		vehicle:removeFromPhysics()
	end

	superFunc(self)
end